#!/usr/bin/env python

# stdlib imports
import pathlib
import re

import numpy as np

# third party imports
import pandas as pd

# local imports
from esi_utils_pager.country import Country

DEFAULT_RATE = 1.17 / 100.0


def adjust_pop(population, tpop, tevent, rate):
    """Adjust input population between two input years given growth rate.

    :param population:
      Population starting value at time *tpop*.
    :param tpop:
      Year in which input population data was collected.
    :param tevent:
      Year to which population data should be adjusted.
    :param rate:
      Population growth rate value.
    :returns:
      Adjusted population value at time *tevent*.
    """
    T = tpop - tevent
    adjpop = np.round(population * np.power((1 + rate), (-1 * T)))
    return adjpop


class PopulationGrowth(object):
    def __init__(self, ratedict, default_rate=DEFAULT_RATE):
        """Initialize Population growth with dictionary containing rates over given time
        spans, per country.

        :param ratedict:
          dictionary like: {841: {'end': [1955, 1960, 1965],
                                  'rate': [0.01, 0.02, 0.03],
                                  'start': [1950, 1955, 1960]},
                            124: {'end': [1955, 1960, 1965],
                                  'rate': [0.02, 0.03, 0.04],
                                  'start': [1950, 1955, 1960]}}
          Where 841 and 842 in this case are country codes (US and Canada), and the three "columns" for each
          country are the year start of each time interval, the year end of each time interval, and the growth
          rates for those time intervals.
        :param default_rate:
          Value to be used for growth rate when input country codes are not found in ratedict.
        """
        # check the fields in the ratedict
        for key, value in ratedict.items():
            if "start" not in value or "end" not in value or "rate" not in value:
                raise Exception(
                    'All country rate dictionaries must contain keys "start","end","rate"'
                )
            if not (len(value["start"]) == len(value["end"]) == len(value["rate"])):
                raise Exception(
                    "Length of start/end year arrays must match length of rate arrays."
                )
        self._dataframe = pd.DataFrame(ratedict)
        self._default = default_rate

    @classmethod
    def fromDefault(cls):
        homedir = pathlib.Path(__file__).parent  # where is this module?
        excelfile = homedir / "data" / "WPP2015_POP_F02_POPULATION_GROWTH_RATE.xls"
        return cls.fromUNSpreadsheet(excelfile)

    @classmethod
    def fromUNSpreadsheet(cls, excelfile, default_rate=DEFAULT_RATE):
        """Instantiate population growth rates from UN global spreadsheet.
        http://esa.un.org/unpd/wpp/Download/Standard/Population/

        :param excelfile:
          Path to Excel file containing UN population growth rate data per country.
        :param default_rate:
          Value to be used for growth rate when input country codes are not found in ratedict.
        :returns:
          PopulationGrowth instance.
        """
        re_year = "[0-9]*"
        df = pd.read_excel(excelfile, header=16)
        ratedict = {}
        starts = []
        ends = []
        for col in df.columns:
            matches = re.findall(re_year, col)
            if len(matches) and len(matches[0]):
                starts.append(int(matches[0]))
                ends.append(int(matches[2]))

        ccode_idx = df.columns.get_loc("Country code")
        uscode = 840
        usrates = None
        country = Country()
        for idx, row in df.iterrows():
            key = row["Country code"]
            rates = row.iloc[ccode_idx + 1 :].values / 100.0
            if key == uscode:
                usrates = rates.copy()
            if country.getCountry(key) is None:
                continue
            ratedict[key] = {"start": starts[:], "end": ends[:], "rate": rates}

        # we have three non-standard "country" codes for California, eastern US, and western US.
        ratedict[902] = {"start": starts[:], "end": ends[:], "rate": usrates}
        ratedict[903] = {"start": starts[:], "end": ends[:], "rate": usrates}
        ratedict[904] = {"start": starts[:], "end": ends[:], "rate": usrates}

        return cls(ratedict, default_rate=default_rate)

    def getRate(self, ccode, year):
        """Return population growth rate(s) for a given country code and year.

        :param ccode:
          Numeric country code.
        :param year:
          Integer year to be used to find growth rate (will be between start and end years,
          or before first start year or after last end year).
        :returns:
          Scalar growth rate.
        """
        ccode = int(ccode)
        if ccode not in self._dataframe.columns:
            return self._default
        starts = np.array(self._dataframe[ccode]["start"])
        ends = np.array(self._dataframe[ccode]["end"])
        rates = np.array(self._dataframe[ccode]["rate"])
        if year is None:
            return dict(list(zip(starts, rates)))
        if year < starts.min():
            rate = rates[0]
        elif year > ends.max():
            rate = rates[-1]
        else:
            idx = (np.abs(year - ends)).argmin()
            rate = rates[idx]
        return rate

    def getRates(self, ccode):
        """Return population growth rates for a given country code.

        :param ccode:
          Numeric country code.
        :param year:
          Integer year to be used to find growth rate (will be between start and end years,
          or before first start year or after last end year).
        :returns:
          Tuple of two lists of (start_years,rates).
        """
        if ccode not in self._dataframe.columns:
            raise Exception(
                "Country %s not found in PopulationGrowth data structure." % ccode
            )
        starts = np.array(self._dataframe[ccode]["start"])
        rates = np.array(self._dataframe[ccode]["rate"])
        return (starts, rates)

    def adjustPopulation(self, population, ccode, tpop, tevent):
        """Adjust population based on growth rates.

        :param population:
          Number of people.
        :param ccode:
          Numeric country code.
        :param tpop:
          Year of population data collection.
        :param tevent:
          Year to which population data should be adjusted from tpop.
        :returns:
          Population adjusted for growth rates in years between tpop and tevent.
        """
        if tpop == tevent:
            return population
        if tpop < tevent:
            interval = 1
        else:
            interval = -1
        newpop = population
        for startpop in np.arange(tpop, tevent, interval):
            endpop = startpop + interval
            rate = self.getRate(ccode, startpop)
            newpop = adjust_pop(newpop, startpop, endpop, rate)

        return newpop



#!/usr/bin/env python

# stdlib imports
import pathlib

import numpy as np

# third party imports
import pandas as pd


class Country(object):
    def __init__(self):
        homedir = pathlib.Path(__file__).parent  # where is this script?
        excelfile = homedir / "data" / "countries.xlsx"
        self._loadFromExcel(excelfile)

    def getUSCode(self, code):
        """Handle US-region specific codes for California, Eastern US, and Western US.

        :param code:
          Input numeric country code.
        :returns:
          840 if input is in (902,903,904), input ccode otherwise.
        """
        if code in (902, 903, 904):
            return 840
        else:
            return code

    def _loadFromExcel(self, excelfile):
        self._dataframe = pd.read_excel(excelfile)
        idx = pd.isnull(self._dataframe["Name"])
        self._dataframe.loc[idx, "Name"] = self._dataframe["LongName"][idx]

    def _loadFromCSV(self, csvfile):
        """Load from a CSV file containing five columns: LongName,ISO2,ISO3,ISON,Name.

        NB:

          There are three PAGER-specific country codes in the PAGER countries.csv file.
          - 902: California
          - 903: Eastern US
          - 904: Western US

        :param csvfile:
          CSV file containing country information with the following columns:
           - LongName: Long country name.
           - ISO2: Two letter country code.
           - ISO3: Three letter country code.
           - ISON: Numeric country code.
           - Name: Short country name.
           - Source: https://en.wikipedia.org/wiki/ISO_3166-1

        """
        self._dataframe = pd.read_csv(csvfile)
        cols = ["LongName", "ISO2", "ISO3", "ISON", "Name"]
        for col in cols:
            if col not in self._dataframe.columns:
                raise Exception(
                    "PAGER country CSV file must contain columns: %s." % str(cols)
                )

        idx = pd.isnull(self._dataframe["Name"])
        self._dataframe.loc[idx, "Name"] = self._dataframe["LongName"][idx]

    def getCountry(self, value):
        """Return a dictionary containing the country name/codes for the first country that matches the input value.

        N.B.

          This method matches the first country it can, which means that in the case of an input "guinea", for
          example, you would get back the data for "Equatorial Guinea", as opposed to "Guinea", "Guinea-Bissau", or
          "Papua New Guinea".  Also be aware that there is a country code for "United States Minor Outlying Islands".

        :param value:
          One of:
           - Two letter ISO 3166 country code (JP,US, etc.)
           - Three letter ISO 3166 country code (JPN,USA, etc.)
           - Numerical ISO 3166 country code (392,840, etc.)
           - Name or name fragment (Japan,united states of america, etc.)

        :returns:
          Dictionary containing the following:
            - Name: Short name (i.e., Bolivia)
            - LongName: Long name (i.e., Plurinational State of Bolivia)
            - ISO2: Two letter ISO country code (BO)
            - ISO3: Two letter ISO country code (BOL)
            - ISON: Numeric ISO country code (68)
            - Population Number of people inside the country.
          or None if the input value does not match any known country data.

        """
        emptyrow = {
            "Name": "Unknown",
            "LongName": "Unknown",
            "ISO2": "UK",
            "ISO3": "UKN",
            "ISON": 0,
            "Population": 0,
        }
        row = None
        if isinstance(value, (int, float, np.number)):
            row = self._dataframe[self._dataframe["ISON"] == value]
        elif isinstance(value, str):
            if len(value) == 0:
                return emptyrow
            if len(value) == 2:
                row = self._dataframe[self._dataframe["ISO2"] == value]
            elif len(value) == 3:
                row = self._dataframe[self._dataframe["ISO3"] == value]
            else:
                row = self._dataframe[
                    self._dataframe.Name.str.lower().str.contains(value.lower())
                ]
        if row is None:
            return emptyrow
        else:
            if len(row) > 0:
                return row.iloc[0].to_dict()
            elif not len(row):
                return emptyrow
            else:
                return row.to_dict()



# stdlib imports
import os.path
import shutil

# third party imports
import yaml


def read_config():
    """Read in configuration parameters from config .py file.

    returns:
      Dictionary containing configuration parameters.
    raises:
      Exception if config file does not exist.
    """
    # get config file name, make sure it exists
    configfilename = get_config_file()
    if configfilename is None:
        raise Exception("Config file could not be found at %s." % configfilename)

    config = yaml.safe_load(open(configfilename, "rt"))
    return config


def read_mail_config():
    """Read in configuration parameters from config .py file.

    returns:
      Dictionary containing configuration parameters.
    raises:
      Exception if config file does not exist.
    """
    # get config file name, make sure it exists
    configfilename = get_mail_config_file()
    if configfilename is None:
        raise Exception("Config file could not be found at %s." % configfilename)

    config = yaml.safe_load(open(configfilename, "rt"))
    return config


def write_config(config, make_backup=True):
    """Write out config parameters.

    :param config:
      Dictionary with configuration parameters.
    :param make_backup:
      Boolean indicating whether a backup of the current config file
      should be made before writing new one.
    """
    # get config file name, make sure it exists
    configfilename = os.path.join(os.path.expanduser("~"), ".losspager", "config.yml")
    if not os.path.isfile(configfilename):
        raise Exception("Config file could not be found at %s." % configfilename)
    backup_name = os.path.join(os.path.expanduser("~"), ".losspager", "config.yml.bak")
    shutil.copyfile(configfilename, backup_name)
    f = open(configfilename, "wt")
    f.write(yaml.dump(config))
    f.close()


def get_mail_config_file():
    """Find and return config file name, if exists.  None returned if file does not exist.

    :returns:
      config file name, or None if config file does not exist.
    """
    configfilename = os.path.join(
        os.path.expanduser("~"), ".losspager", "mailconfig.yml"
    )
    if not os.path.isfile(configfilename):
        return None
    return configfilename


def get_config_file():
    """Find and return config file name, if exists.  None returned if file does not exist.

    :returns:
      config file name, or None if config file does not exist.
    """
    configfilename = os.path.join(os.path.expanduser("~"), ".losspager", "config.yml")
    if not os.path.isfile(configfilename):
        return None
    return configfilename



#!/usr/bin/env python

# stdlib imports
import pathlib
import re

# third party imports
import numpy as np
import pandas as pd
from mapio.grid2d import Grid2D

# local imports
from esi_utils_pager.country import Country
from esi_utils_pager.emploss import EmpiricalLoss
from esi_utils_pager.exposure import Exposure
from esi_utils_pager.growth import PopulationGrowth

GLOBAL_GDP = 16100  # from https://en.wikipedia.org/wiki/Gross_world_product


class GDP(object):
    def __init__(self, dataframe):
        """Create an instance of a GDP object with a dataframe of countries/GDP values over time.

        :param dataframe:
          Pandas dataframe where rows are countries, and columns are rates for different years.
          dataframe should look like this:
          DataFrame({'Country Code':['AFG'],
                     '1960':[59.78768071],
                     '1961':[59.89003694],})
          The above contains the GDP values for Afghanistan for 1960 and 196
        """
        self._dataframe = dataframe
        self._country = Country()

    @classmethod
    def fromDefault(cls):
        homedir = pathlib.Path(__file__).parent  # where is this module?
        excelfile = homedir / "data" / "API_NY.GDP.PCAP.CD_DS2_en_excel_v2_4770383.xls"
        return cls.fromWorldBank(excelfile)

    @classmethod
    def fromWorldBank(cls, excelfile):
        """Read in Excel data from the World Bank containing per capita GDP values for all countries in the world.
        Taken from: http://data.worldbank.org/indicator/NY.GDP.PCAP.CD

        :param excelfile:
          Excel spreadsheet downloaded from above source.
        :returns:
          GDP instance.
        """
        df = pd.read_excel(excelfile, sheet_name="Data", header=3)
        return cls(df)

    def getGDP(self, ccode, year):
        """Get the GDP value for a given country code and a particular year.

        :param ccode:
          Any of ISO2, ISO3, or ISON country codes.
        :param year:
          Year of desired GDP value.  If this year is before the earliest year in the data source,
          the earliest GDP value will be used. If this year is after the latest year in the data source,
          the latest non-NaN GDP value will be used.
        :returns:
          Tuple of:
            - GDP value corresponding to country code and year, unless country code is not found, in which case
            a default global GDP value will be returned.
            - The country code which is most applicable to the output GDP.  For example, if the GDP value chosen
              is the global value, then this country code will be None.  If the input country code is XF (California),
              then the output country code will be 'US'.
        """
        # first make sure the ccode is valid...
        countrydict = self._country.getCountry(ccode)
        if countrydict is None:
            return (GLOBAL_GDP, None)
        if countrydict["ISO2"] in ["XF", "EU", "WU"]:
            ccode = "USA"
            outccode = "US"
        else:
            ccode = countrydict["ISO3"]
            outccode = ccode

        if not self._dataframe["Country Code"].isin([ccode]).any():
            return (GLOBAL_GDP, None)

        row = self._dataframe.loc[self._dataframe["Country Code"] == ccode].iloc[0]
        columns = row.index.tolist()
        str_years = [c for c in columns if re.match("[0-9]{4}", c) is not None]
        gdp_values = np.array(row[str_years].values, dtype=np.float32)
        years = np.array([int(y) for y in str_years])

        # extract finite values from gdp array and years
        realidx = np.isfinite(gdp_values)
        gdp_values = gdp_values[realidx]
        years = years[realidx]
        if not len(years):
            return (GLOBAL_GDP, None)
        if year < years.min():
            yearidx = years.argmin()
        elif year > years.max():
            yearidx = years.argmax()
        else:
            if year in years:
                yearidx = np.where(years == year)[0][0]
            else:
                yearidx = None
        if yearidx is not None:
            gdp = gdp_values[yearidx]
        else:
            gdp = np.interp(year, years, gdp_values)

        return (gdp, outccode)


class EconExposure(Exposure):
    def __init__(self, popfile, popyear, isofile):
        """Create instance of EconExposure class (subclass of Exposure, and shares methods of that class.)

        :param popfile:
          Any GMT or ESRI style grid file supported by MapIO, containing population data.
        :param popyear:
          Integer indicating year when population data is valid.
        :param isofile:
          Any GMT or ESRI style grid file supported by MapIO, containing country code data (ISO 3166-1 numeric).
        """
        self._emploss = EmpiricalLoss.fromDefaultEconomic()
        self._gdp = GDP.fromDefault()
        self._econpopgrid = None
        popgrowth = PopulationGrowth.fromDefault()
        super(EconExposure, self).__init__(popfile, popyear, isofile)

    def getEconPopulationGrid(self):
        """Return the internal economic exposure population grid, created during calcExposure().

        :raises:
          PagerException when calcExposure() has not been called.
        """
        if self._econpopgrid is None:
            raise PagerException(
                "Must call calcExposure() before calling getEconPopulationGrid()."
            )
        return self._econpopgrid

    def calcExposure(self, shakefile):
        """Calculate population exposure to shaking.

        Calculate population exposure to shaking, per country, multiplied by event-year per-capita GDP and
        alpha correction factor.  Also multiply the internal population grid by GDP and
        alpha.

        :param shakefile:
          Path to ShakeMap grid.xml file.
        :returns:
          Dictionary containing country code (ISO2) keys, and values of
          10 element arrays representing population exposure to MMI 1-10.
          Dictionary will contain an additional key 'Total', with value of exposure across all countries.
        """
        # create a copy of population grid to hold population * gdp * alpha
        expdict = super(EconExposure, self).calcExposure(shakefile)
        self._econpopgrid = Grid2D.copyFromGrid(self._popgrid)
        econdict = {}
        isodata = self._isogrid.getData()
        eventyear = self.getShakeGrid().getEventDict()["event_timestamp"].year
        total = np.zeros((10,))
        for ccode, exparray in expdict.items():
            if ccode.find("Total") > -1 or ccode.find("maximum") > -1:
                continue
            if ccode == "UK":  # unknown
                continue
            lossmodel = self._emploss.getModel(ccode)
            gdp, outccode = self._gdp.getGDP(ccode, eventyear)
            isocode = self._country.getCountry(ccode)["ISON"]
            alpha = lossmodel.alpha
            econarray = exparray * gdp * alpha
            cidx = isodata == isocode
            # multiply the population grid by GDP and alpha, so that when the loss model
            # queries the grid later, those calculations don't have to be re-done.
            self._econpopgrid._data[cidx] = self._econpopgrid._data[cidx] * gdp * alpha
            econdict[ccode] = econarray
            total += econarray

        econdict["TotalEconomicExposure"] = total
        return econdict



#!/usr/bin/env python

# stdlib
import warnings

# third party
import numpy as np
from mapio.reader import get_file_geodict, read
from mapio.shake import ShakeGrid

# local imports
from esi_utils_pager.country import Country
from esi_utils_pager.growth import PopulationGrowth

SCENARIO_WARNING = (
    10  # number of years after date of population data to issue a warning
)
SCENARIO_ERROR = (
    20  # number of years after date of population data to raise an exception
)


def calc_exposure(mmidata, popdata, isodata):
    """Calculate population exposure to shaking per country.

    :param mmidata:
      Scalar or array-like value containing floating point MMI data values (range 1-10).
    :param popdata:
      Scalar or array-like value containing integer population data values.
    :param isodata:
      Scalar or array-like value containing integer country-code (ISO 3166-1 numeric) data values.
    :returns:
      Dictionary of population exposures to shaking, keys are country code, values are 10-element arrays.
    """
    mmidata = np.array(mmidata)
    popdata = np.array(popdata)
    isodata = np.array(isodata)
    exposures = {}
    ccodes = np.unique(isodata[~np.isnan(isodata)])
    for ccode in ccodes:
        cidx = np.ravel_multi_index(np.where(isodata == ccode), isodata.shape)
        expsum = np.zeros((10), dtype=np.uint32)
        for mmi in range(1, 11):
            mmi_lower = mmi - 0.5
            mmi_upper = mmi + 0.5
            midx = np.ravel_multi_index(
                np.where((mmidata >= mmi_lower) & (mmidata < mmi_upper)), mmidata.shape
            )
            idx = np.unravel_index(np.intersect1d(cidx, midx), mmidata.shape)
            popsum = np.nansum(popdata[idx])
            expsum[mmi - 1] = int(popsum)
        exposures[ccode] = expsum[:]

    return exposures


class Exposure(object):
    def __init__(self, popfile, popyear, isofile, popgrowth=None):
        """Create Exposure object, with population and country code grid files,
        and a dictionary of country growth rates.

        :param popfile:
          Any GMT or ESRI style grid file supported by MapIO, containing population data.
        :param popyear:
          Integer indicating year when population data is valid.
        :param isofile:
          Any GMT or ESRI style grid file supported by MapIO, containing country code data (ISO 3166-1 numeric).
        """
        self._popfile = popfile
        self._popyear = popyear
        self._isofile = isofile
        self._popgrid = None
        self._isogrid = None
        self._shakegrid = None
        if popgrowth is not None:
            self._popgrowth = popgrowth
        else:
            self._popgrowth = PopulationGrowth.fromDefault()
        self._country = Country()

    def calcExposure(self, shakefile):
        """Calculate population exposure to shaking, per country, plus total exposure across all countries.

        :param shakefile:
          Path to ShakeMap grid.xml file.
        :returns:
          Dictionary containing country code (ISO2) keys, and values of
          10 element arrays representing population exposure to MMI 1-10.
          Dictionary will contain an additional key 'TotalExposure', with value of exposure across all countries.
          Dictionary will also contain a field "maximum_border_mmi" which indicates the maximum MMI value along
          any edge of the ShakeMap.
        """
        # get shakemap geodict
        shakedict = ShakeGrid.getFileGeoDict(shakefile, adjust="res")

        # get population geodict
        popdict = get_file_geodict(self._popfile)

        # get country code geodict
        isodict = get_file_geodict(self._isofile)

        # special case for very high latitude events that may be outside the bounds
        # of our population data...
        if not popdict.intersects(shakedict):
            expdict = {"UK": np.zeros((10,)), "TotalExposure": np.zeros((10,))}
            return expdict

        if popdict == shakedict == isodict:
            # special case, probably for testing...
            self._shakegrid = ShakeGrid.load(shakefile, adjust="res")
            self._popgrid = read(self._popfile)
            self._isogrid = read(self._isofile)
        else:
            sampledict = popdict.getBoundsWithin(shakedict)
            self._shakegrid = ShakeGrid.load(
                shakefile,
                samplegeodict=sampledict,
                resample=True,
                method="linear",
                adjust="res",
            )
            self._popgrid = read(
                self._popfile,
                samplegeodict=sampledict,
                resample=False,
                doPadding=True,
                padValue=np.nan,
            )
            self._isogrid = read(
                self._isofile,
                samplegeodict=sampledict,
                resample=True,
                method="nearest",
                doPadding=True,
                padValue=0,
            )

        mmidata = self._shakegrid.getLayer("mmi").getData()
        popdata = self._popgrid.getData()
        isodata = self._isogrid.getData()

        eventyear = self._shakegrid.getEventDict()["event_timestamp"].year

        # in order to avoid crazy far-future scenarios where PAGER models are probably invalid,
        # check to see if the time gap between the date of population data collection and event year
        # reaches either of a couple of different thresholds.
        if eventyear > self._popyear:
            tdiff = eventyear - self._popyear
            if tdiff > SCENARIO_WARNING and tdiff < SCENARIO_ERROR:
                msg = (
                    """The input ShakeMap event year is more than %i years from the population date.
                PAGER results for events this far in the future may not be valid."""
                    % SCENARIO_WARNING
                )
                warnings.warn(msg)
            if tdiff > SCENARIO_ERROR:
                msg = (
                    """The input ShakeMap event year is more than %i years from the population date.
                PAGER results for events this far in the future are not valid. Stopping."""
                    % SCENARIO_ERROR
                )
                raise Exception(msg)

        ucodes = np.unique(isodata[~np.isnan(isodata)])
        for ccode in ucodes:
            cidx = isodata == ccode
            popdata[cidx] = self._popgrowth.adjustPopulation(
                popdata[cidx], ccode, self._popyear, eventyear
            )

        exposure_dict = calc_exposure(mmidata, popdata, isodata)
        newdict = {}
        # Get rolled up exposures
        total = np.zeros((10,), dtype=np.uint32)
        for isocode, value in exposure_dict.items():
            cdict = self._country.getCountry(int(isocode))
            if cdict is None:
                ccode = "UK"
            else:
                ccode = cdict["ISO2"]
            newdict[ccode] = value
            total += value

        newdict["TotalExposure"] = total

        # get the maximum MMI value along any of the four map edges
        nrows, ncols = mmidata.shape
        top = mmidata[0, 0:ncols].max()
        bottom = mmidata[nrows - 1, 0:ncols].max()
        left = mmidata[0:nrows, 0].max()
        right = mmidata[0:nrows, ncols - 1].max()
        newdict["maximum_border_mmi"] = np.array([top, bottom, left, right]).max()

        return newdict

    def getPopulationGrid(self):
        """Return the internal population grid.

        :returns:
          Population grid.
        """
        if self._popgrid is None:
            raise Exception("calcExposure() method must be called first.")
        return self._popgrid

    def getCountryGrid(self):
        """Return the Grid2D object containing ISO numeric country codes.

        :returns:
          Grid2D object containing ISO numeric country codes.
        """
        if self._isogrid is None:
            raise Exception("calcExposure() method must be called first.")
        return self._isogrid

    def getShakeGrid(self):
        """Return the MultiGrid object containing ShakeMap data.

        :returns:
          MultiGrid object containing ShakeMap data.
        """
        if self._shakegrid is None:
            raise Exception("calcExposure() method must be called first.")
        return self._shakegrid



import numpy as np
import math
from scipy.special import erfc, erfcinv


def phi(input):
    """Phi function.

    :param input:
      Float (scalar or array) value.
    :returns:
      phi(input).
    """
    return 0.5 * erfc(-input / np.sqrt(2))


def invphi(input):
    """Inverse phi function.

    :param input:
      Float (scalar or array) value.
    :returns:
      invphi(input).
    """
    return -1 * np.sqrt(2) * erfcinv(input / 0.5)


def calcEmpiricalProbFromValue(G, e, value):
    """Calculate the empirical probability of a given value of loss (fatalities, dollars).

    :param G:
      Input G statistic.
    :param e:
      Expected number of losses.
    :param value:
      Number of losses for which corresponding probability should be calculated.
    :returns:
      Probability that value will not be exceeded.
    """
    e = e + 0.00001
    value = value + 0.00001
    p = phi((math.log(value) - math.log(e)) / G)
    return p


def calcEmpiricalValueFromProb(G, e, p):
    """Calculate the loss value given an input probability.

    :param G:
      Input G statistic.
    :param e:
      Expected number of losses.
    :param p:
      Input probability for which corresponding loss value should be calculated.
    :returns:
      Number of losses.
    """
    e = e + 0.00001
    value = math.exp(G * invphi(p) + math.log(e))
    return value


def calcEmpiricalProbFromRange(G, e, drange):
    """Calculate the empirical probability of a given loss range.

    :param G:
      Input G statistic.
    :param e:
      Expected number of losses.
    :param drange:
      Two-element sequence, containing range of losses over which probability should be calculated.
    :returns:
      Probability that losses will occur in input range.
    """
    e = e + 0.00001
    if e < 1:
        # e = 0.1
        if G > 1.7:
            G = 1.7613
    if len(drange) == 2:
        if (e - 0) < 0.001:
            e = 0.5
        fmin = drange[0] + 0.00001
        fmax = drange[1] + 0.00001
        p1 = phi((math.log(fmax) - math.log(e)) / G)
        p2 = phi((math.log(fmin) - math.log(e)) / G)
        p = p1 - p2
        return p
    else:
        psum = 0
        for i in range(0, len(drange) - 1):
            fmax = drange[i + 1] + 0.00001
            fmin = drange[i] + 0.00001
            p1 = phi((math.log(fmax) - math.log(e)) / G)
            p2 = phi((math.log(fmin) - math.log(e)) / G)
            psum += p1 - p2
        return psum



#!/usr/bin/env python

# stdlib imports
import argparse
import json
import logging
import pathlib
import sys
import tempfile
from urllib.request import urlopen

# third party imports
import numpy as np
import pandas as pd
import rasterio

# local imports
from esi_utils_pager.calc import calc_pager_event, calc_pager_events
from esi_utils_pager.config import read_config

EVENT_TEMPLATE = (
    "https://earthquake.usgs.gov/earthquakes/feed/v1.0/detail/{eventid}.geojson"
)


def main():
    helpstr = (
        "Render complete empirical/semi-empirical PAGER results.\n\n"
        "Default behavior renders PAGER results for a set of earthquakes\n"
        "as a formatted DataFrame with multiple rows of exposure and loss,\n"
        "one row per country, plus a final row with totals. Output columns (in order):\n"
        "EventID: ComCat event ID\n"
        "Time: Event Time (y-m-d h:m:s)\n"
        "Latitude: Event hypocentral latitude\n"
        "Longitude: Event hypocentral longitude\n"
        "Depth: Event hypocentral depth\n"
        "Magnitude: Event magnitude\n"
        "Location: Event location description\n"
        "CountryCode: Country code where exposures/losses occur (or Total)\n"
        "MMI01: Population exposure to shaking at MMI level 1\n"
        "...\n"
        "MMI10: Population exposure to shaking at MMI level 10\n"
        "Fatalities: Fatalities due to shaking\n"
        "EconMMI01: Economic exposure to shaking at MMI level 1\n"
        "...\n"
        "EconMMI10: Economic exposure to shaking at MMI level 10\n"
        "Dollars: Economic losses (USD) due to shaking\n"
    )
    parser = argparse.ArgumentParser(
        description=helpstr,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "-f", "--folder", help="A folder containing many ShakeMap grid.xml files."
    )
    group.add_argument("-g", "--grid-xml", help="A ShakeMap grid.xml file.")
    group.add_argument(
        "-e",
        "--eventid",
        help="ComCat event ID - preferred ShakeMap grid.xml will be used.",
    )
    parser.add_argument(
        "-s",
        "--semimodel",
        help="Calculate semi-empirical model results as well.",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "-o", "--outfile", help="Specify output file (.xlsx for Excel, .csv for CSV)"
    )
    parser.add_argument(
        "-v",
        "--verbose",
        default=False,
        action="store_true",
        help="Print progress output to the screen",
    )

    args = parser.parse_args()

    if args.verbose:
        loglevel = logging.INFO
    else:
        loglevel = logging.CRITICAL
    logging.basicConfig(
        level=loglevel,
        format="%(asctime)s %(message)s",
        handlers=[logging.StreamHandler()],
    )

    if args.folder:
        dataframe = calc_pager_events(args.folder, args.verbose, args.semimodel)
    elif args.eventid:
        url = EVENT_TEMPLATE.format(eventid=args.eventid)
        with tempfile.TemporaryDirectory() as tempdir:
            with urlopen(url) as fh:
                data = fh.read().decode("utf8")
                jdict = json.loads(data)
                if "shakemap" not in jdict["properties"]["types"]:
                    print(f"No ShakeMap for event {args.eventid}. Exiting.")
                    sys.exit(1)
                shakemap = jdict["properties"]["products"]["shakemap"][0]
                grid_url = shakemap["contents"]["download/grid.xml"]["url"]
                with urlopen(grid_url) as fh2:
                    xdata = fh2.read().decode("utf8")

                tmpgridfile = pathlib.Path(tempdir) / "tmp.xml"
                with open(tmpgridfile, "wt") as fout:
                    fout.write(xdata)
                config = read_config()
                dataframe = calc_pager_event(tmpgridfile, config, args.semimodel)
    else:
        config = read_config()
        dataframe = calc_pager_event(args.grid_xml, config, args.semimodel)

    if args.outfile:
        print(f"Saving {len(dataframe)} rows to {args.outfile}")
        if args.outfile.endswith(".xlsx"):
            dataframe.to_excel(args.outfile, index=False)
        else:
            dataframe.to_csv(args.outfile, index=False)
        sys.exit(0)
    else:
        print(dataframe.to_string(index=False))


if __name__ == "__main__":
    main()



#!/usr/bin/env python

# stdlib imports
import pathlib
from collections import OrderedDict
from xml.dom import minidom

# third party imports
import numpy as np
import shapely
from mapio.grid2d import Grid2D
from scipy.special import erfc

# local imports
from esi_utils_pager.country import Country
from esi_utils_pager.probs import calcEmpiricalProbFromRange

DEFAULT_THETA = 16.0
DEFAULT_BETA = 0.15
DEFAULT_L2G = 1.0
DEFAULT_ALPHA = 1.0


class LossModel(object):
    def __init__(self, name, rates, l2g, alpha=None):
        """Create a loss model from an array of loss rates at MMI 1-10.

        :param name:
          Name (usually two letter country code) for model.
        :param rates:
          Array-like float values 10 elements in length.
        :param l2g:
          Float value defining the value of the L2G norm calculated when model was derived.
        :param alpha:
          Float value defining the alpha (economic correction factor) value for the model.
          Not specified/used for fatality models.
        :returns:
          LossModel instance.
        """
        self._name = name
        self._rates = rates[:]
        self._l2g = l2g
        self._alpha = alpha

    def __repr__(self):
        """return string representation of loss model."""
        mmirange = np.arange(5, 10)
        rates = self.getLossRates(mmirange)
        reprstr = ""
        for i in range(0, len(mmirange)):
            mmi = mmirange[i]
            rate = rates[i]
            reprstr += "MMI %i: 1 in %s\n" % (mmi, format(int(1.0 / rate), ",d"))
        return reprstr

    def getLossRates(self, mmirange):
        mmirange = np.array(mmirange)
        idx = mmirange - 1
        return self._rates[idx]

    @property
    def name(self):
        """Return the name associated with this model.

        :returns:
          The name associated with this model.
        """
        return self._name

    @property
    def theta(self):
        """Return the theta value associated with this model.

        :returns:
          The theta associated with this model.
        """
        return self._theta

    @property
    def beta(self):
        """Return the beta value associated with this model.

        :returns:
          The beta value associated with this model.
        """
        return self._beta

    @property
    def alpha(self):
        """Return the alpha value associated with this model (may be None).

        :returns:
          The alpha value associated with this model (may be None).
        """
        return self._alpha

    @property
    def l2g(self):
        """Return the L2G value associated with this model.

        :returns:
          The L2G value associated with this model.
        """
        return self._l2g

    def getLosses(self, exp_pop, mmirange, rates=None):
        """Calculate losses given input arrays of population exposures and MMI values.

        :param exp_pop:
          Array of population exposed at mmirange values.
        :param mmirange:
          Array of MMI values exp_pop is exposed to.
        :param rates:
          Array of 10 loss rates which, if specified, will be used instead of the lognormal rates.
        :returns:
          Scalar floating point number of losses.
        """
        if rates is None:
            rates = self.getLossRates(mmirange)
        deaths = np.nansum(rates * exp_pop)
        return deaths

    def getArea(self):
        """Calculate the area under the loss rate curve (defined for MMI 5-9).

        Used internally for model to model comparisons.

        :returns:
          Area under the loss rate curve (defined for MMI 1-10).
        """
        mmirange = np.arange(5, 10)
        rates = self.getLossRates(mmirange)
        area = np.trapz(rates, mmirange)
        return area

    def __lt__(self, other):
        """Is this model less deadly than other model?

        :param other:
          Another LognormalModel instance.
        :returns:
          True if this model is less deadly than other model.
        """
        area1 = self.getArea()
        area2 = other.getArea()
        if area1 < area2:
            return True

    def __le__(self, other):
        """Is this model less than or just as deadly as other model?

        :param other:
          Another LognormalModel instance.
        :returns:
          True if this model is less than or just as deadly as other model.
        """
        area1 = self.getArea()
        area2 = other.getArea()
        if area1 <= area2:
            return True

    def __eq__(self, other):
        """Is this model equally deadly as other model?

        :param other:
          Another LognormalModel instance.
        :returns:
          True if this model is equally deadly as other model.
        """
        area1 = self.getArea()
        area2 = other.getArea()
        if area1 == area2:
            return True

    def __gt__(self, other):
        """Is this model more deadly than other model?

        :param other:
          Another LognormalModel instance.
        :returns:
          True if this model is more deadly than other model.
        """
        area1 = self.getArea()
        area2 = other.getArea()
        if area1 > area2:
            return True

    def __ge__(self, other):
        """Is this model greater than or just as deadly as other model?

        :param other:
          Another LognormalModel instance.
        :returns:
          True if this model is greater than or just as deadly as other model.
        """
        area1 = self.getArea()
        area2 = other.getArea()
        if area1 >= area2:
            return True


class LoglinearModel(LossModel):
    """Loglinear loss model (defined by theta/beta (or mu/sigma) values."""

    def __init__(self, name, theta, beta, l2g, alpha=None):
        """Instantiate Loglinear Loss object.

        :param name:
          Name (usually two letter country code) for model.
        :param theta:
          Float value defining the theta (or mu) value for the model.
        :param beta:
          Float value defining the beta (or sigma) value for the model.
        :param l2g:
          Float value defining the value of the L2G norm calculated when model was derived.
        :param alpha:
          Float value defining the alpha (economic correction factor) value for the model.
          Not specified/used for fatality models.
        :returns:
          LognormalModel instance.
        """
        self._name = name
        self._theta = theta
        self._beta = beta
        self._l2g = l2g
        self._alpha = alpha

    def getLossRates(self, mmirange):
        """Get the loss rates at each of input MMI values.

        :param mmirange:
          Array-like range of MMI values at which loss rates will be calculated.
        :returns:
          Array of loss rates for input MMI values.
        """
        mmi = np.array(mmirange)
        yy = numpy.power(10, (theta - (mmi * beta)))
        return yy


class LognormalModel(LossModel):
    """Lognormal loss model (defined by theta/beta (or mu/sigma) values."""

    def __init__(self, name, theta, beta, l2g, alpha=None):
        """Instantiate Lognormal Loss object.

        :param name:
          Name (usually two letter country code) for model.
        :param theta:
          Float value defining the theta (or mu) value for the model.
        :param beta:
          Float value defining the beta (or sigma) value for the model.
        :param l2g:
          Float value defining the value of the L2G norm calculated when model was derived.
        :param alpha:
          Float value defining the alpha (economic correction factor) value for the model.
          Not specified/used for fatality models.
        :returns:
          LognormalModel instance.
        """
        self._name = name
        self._theta = theta
        self._beta = beta
        self._l2g = l2g
        self._alpha = alpha

    def getLossRates(self, mmirange):
        """Get the loss rates at each of input MMI values.

        :param mmirange:
          Array-like range of MMI values at which loss rates will be calculated.
        :returns:
          Array of loss rates for input MMI values.
        """
        mmi = np.array(mmirange)
        xx = np.log(mmirange / self._theta) / self._beta
        yy = 0.5 * erfc(-xx / np.sqrt(2))
        return yy


class EmpiricalLoss(object):
    """Container class for multiple LognormalModel objects."""

    def __init__(self, model_list, losstype="fatality"):
        """Instantiate EmpiricalLoss class.

        :param model_list:
          List of LognormalModel objects.  The names of these will be used as keys for the getModel() method.
        :param losstype:
          One of 'fatality' or 'economic'.
        :returns:
          EmpiricalLoss instance.
        """
        if losstype not in ["fatality", "economic"]:
            raise Exception('losstype must be one of ("fatality","economic").')
        self._loss_type = losstype
        self._model_dict = {}
        for model in model_list:
            self._model_dict[model.name] = model
        self._country = (
            Country()
        )  # object that can translate between different ISO country representations.
        self._overrides = (
            {}
        )  # dictionary of manually set rates (not necessarily lognormal)

    def getModel(self, ccode):
        """Return the LognormalModel associated with given country code,
        or a default model if country code not found.

        :param ccode:
          Usually two letter ISO country code.
        :returns:
          LognormalModel instance containing model for input country code, or a default model.
        """
        ccode = ccode.upper()
        default = LognormalModel(
            "default", DEFAULT_THETA, DEFAULT_BETA, DEFAULT_L2G, alpha=DEFAULT_ALPHA
        )
        if ccode in self._model_dict:
            return self._model_dict[ccode]
        else:
            return default

    @classmethod
    def fromDefaultFatality(cls):
        homedir = pathlib.Path(__file__).parent  # where is this module?
        fatxml = homedir / "data" / "fatality.xml"
        return cls.fromXML(fatxml)

    @classmethod
    def fromDefaultEconomic(cls):
        homedir = pathlib.Path(__file__).parent  # where is this module?
        econxml = homedir / "data" / "economy.xml"
        return cls.fromXML(econxml)

    @classmethod
    def fromXML(cls, xmlfile):
        """Load country-specific models from an XML file of the form:
          <?xml version="1.0" encoding="US-ASCII" standalone="yes"?>

          <models vstr="2.2" type="fatality">

            <model ccode="AF" theta="11.613073" beta="0.180683" gnormvalue="1.0"/>

          </models>

          or

          <?xml version="1.0" encoding="US-ASCII" standalone="yes"?>

          <models vstr="1.3" type="economic">

            <model alpha="15.065400" beta="0.100000" gnormvalue="4.113200" ccode="AF"/>

          </models>

        :param xmlfile:
          XML file containing model parameters (see above).
        :returns:
          EmpiricalLoss instance.
        """
        root = minidom.parse(str(xmlfile))
        rootmodels = root.getElementsByTagName("models")[0]
        models = rootmodels.getElementsByTagName("model")
        losstype = rootmodels.getAttribute("type")
        model_list = []
        for model in models:
            key = model.getAttribute("ccode")
            theta = float(model.getAttribute("theta"))
            beta = float(model.getAttribute("beta"))
            l2g = float(model.getAttribute("gnormvalue"))
            if model.hasAttribute("alpha"):
                alpha = float(model.getAttribute("alpha"))
            else:
                alpha = 1.0  # what is the appropriate default value for this?
            model_list.append(LognormalModel(key, theta, beta, l2g, alpha=alpha))
        root.unlink()

        return cls(model_list, losstype)

    def getLossRates(self, ccode, mmirange):
        """Return loss rates for given country country code model at input MMI values.

        :param ccode:
          Country code (usually two letter ISO code).
        :param mmirange:
          Array-like range of MMI values at which loss rates will be calculated.
        :returns:
          Rates from LognormalModel associated with ccode, or default model (see getModel()).
        """
        # mmirange is mmi value, not index
        model = self.getModel(ccode)
        yy = model.getLossRates(mmirange)
        return yy

    def getLosses(self, exposure_dict):
        """Given an input dictionary of ccode (usually ISO numeric), calculate losses per country and total losses.

        :param exposure_dict:
          Dictionary containing country code keys, and 10 element arrays representing population
          exposures to shaking from MMI values 1-10.  If loss type is economic, then this
          input represents exposure *x* per capita GDP *x* alpha (a correction factor).
        :returns:
          Dictionary containing country code keys and integer population estimations of loss.
        """
        # Get a loss dictionary
        fatdict = {}
        for ccode, exparray in exposure_dict.items():
            # exposure array will now also have a row of Total Exposure to shaking.
            if ccode.find("Total") > -1 or ccode.find("maximum") > -1:
                continue
            if ccode == "UK":  # unknown
                continue
            mmirange = np.arange(5, 10)
            model = self.getModel(ccode)
            if ccode in self._overrides:
                rates = self.getOverrideModel(ccode)[4:9]
            else:
                rates = None

            expo = exparray[:]
            expo[8] += expo[9]
            expo = expo[4:9]  # should now be the same size as rates array
            losses = model.getLosses(expo, mmirange, rates=rates)
            fatdict[ccode] = int(losses)

        # now go through the whole list, and get the total number of losses.
        total = sum(list(fatdict.values()))
        if self._loss_type == "fatality":
            fatdict["TotalFatalities"] = total
        else:
            fatdict["TotalDollars"] = total
        return fatdict

    def getCombinedG(self, lossdict):
        """Get combined L2G statistic for all countries contributing to losses.

        :param lossdict:
          Dictionary (as retued by getLosses() method, containing keys of ISO2 country codes,
          and values of loss (fatalities or dollars) for that country.
        :returns:
          sqrt(sum(l2g^2)) for array of all l2g values from countries that had non-zero losses.
        """
        # combine g norm values from all countries that contributed to losses, or if NO losses in
        # any countries, then the combined G from all of them.
        g = []
        has_loss = np.sum(list(lossdict.values())) > 0
        for ccode, loss in lossdict.items():
            if has_loss:
                if loss > 0:
                    g.append(self.getModel(ccode).l2g)
            else:
                g.append(self.getModel(ccode).l2g)

        g = np.array(g)
        zetf = np.sqrt(np.sum(np.power(g, 2)))
        if zetf > 2.5:
            zetf = 2.5
        return zetf

    def getProbabilities(self, lossdict, G):
        """Calculate probabilities over the standard PAGER loss ranges.

        :param lossdict:
          Dictionary (as retued by getLosses() method, containing keys of ISO2 country codes,
          and values of loss (fatalities or dollars) for that country.
        :param G:
          Combined G value (see getCombinedG() method).
        :returns:
          Ordered Dictionary of probability of losses over ranges :
           - '0-1' (green alert)
           - '1-10' (yellow alert)
           - '10-100' (yellow alert)
           - '100-1000' (orange alert)
           - '1000-10000' (red alert)
           - '10000-100000' (red alert)
           - '100000-10000000' (red alert)
        """
        ranges = OrderedDict(
            [
                ("0-1", 0.0),
                ("1-10", 0.0),
                ("10-100", 0.0),
                ("100-1000", 0.0),
                ("1000-10000", 0.0),
                ("10000-100000", 0.0),
                ("100000-10000000", 0.0),
            ]
        )

        if self._loss_type == "economic":
            expected = lossdict["TotalDollars"]
            expected = expected / 1e6  # turn USD into millions of USD
        else:
            expected = lossdict["TotalFatalities"]
        for rangekey, value in ranges.items():
            rparts = rangekey.split("-")
            rmin = int(rparts[0])
            if len(rparts) == 1:
                rmax = int(rparts[0])
            else:
                rmax = int(rparts[1])
                # the high end of the highest red range should be a very large number (ideally infinity).
                # one trillion should do it.
                if rmax == 10000000:
                    rmax = 1e12
            prob = calcEmpiricalProbFromRange(G, expected, (rmin, rmax))
            ranges[rangekey] = prob
        return ranges

    def getAlertLevel(self, lossdict):
        """Get the alert level associated with the input losses.

        :param lossdict:
          Loss results dictionary as returned by getLosses() method.
        :returns:
          String alert level, one of ('green','yellow','orange','red').
        """
        levels = [(1, "green"), (100, "yellow"), (1000, "orange"), (1e12, "red")]
        if "TotalFatalities" in lossdict:
            total = lossdict["TotalFatalities"]
        else:
            total = lossdict["TotalDollars"] / 1e6
        for i in range(0, len(levels) - 1):
            lossmax, thislevel = levels[i]
            if total < lossmax:
                return thislevel
        return "red"  # we should never get here, unless we have 1e18 USD in losses!

    def overrideModel(self, ccode, rates):
        """Override the rates determined from theta,beta values with these hard-coded ones.
        Once set on the instance object, these will be the preferred rates.

        NB: While probably most useful for testing, this method may have real-world uses, so we are
        exposing it in the interface.

        :param ccode:
          (Usually) two-letter ISO country code.
        :param rates:
          10 element (MMI 1-10) array of loss rates which will be used to calculate losses.
        """
        self._overrides[ccode] = rates

    def getOverrideModel(self, ccode):
        """Get the override rates for the input country code.  If not set, None will be returned.

        :param ccode:
          ISO 2 letter country code.
        :returns:
          10 element (MMI 1-10) array of loss rates used to calculate losses for input country code,
          or None.
        """
        if ccode in self._overrides:
            return self._overrides[ccode]
        else:
            return None

    def clearOverrides(self):
        """Clear out any models that have been set manually using overrideModel()."""
        self._overrides.clear()

    def getLossGrid(self, mmidata, popdata, isodata):
        """Calculate floating point losses on a grid.

        :param mmidata:
          Array of MMI values, dimensions (M,N).
        :param popdata:
          Array of population values, dimensions (M,N).
        :param isodata:
          Array of numeric country code values, dimensions (M,N).
        :returns:
          Grid of floating point loss values, dimensions (M,N).
        """
        ucodes = np.unique(isodata)
        fatgrid = np.zeros_like(mmidata)
        # we treat MMI 10 as MMI 9 for modeling purposes...
        mmidata[mmidata > 9.5] = 9.0

        for isocode in ucodes:
            countrydict = self._country.getCountry(int(isocode))
            if countrydict is None:
                ccode = "unknown"
            else:
                ccode = countrydict["ISO2"]

            if ccode not in self._overrides:
                rates = self.getLossRates(ccode, np.arange(5, 10))
            else:
                rates = self.getOverrideModel(ccode)[4:9]

            tcidx = np.where(isodata == isocode)
            cidx = np.ravel_multi_index(tcidx, isodata.shape)
            for i in range(0, len(rates)):
                mmi = i + 5
                mmi_lower = mmi - 0.5
                mmi_upper = mmi + 0.5
                midx = np.ravel_multi_index(
                    np.where((mmidata >= mmi_lower) & (mmidata < mmi_upper)),
                    mmidata.shape,
                )
                idx = np.intersect1d(cidx, midx)
                idx2d = np.unravel_index(idx, mmidata.shape)
                fatgrid[idx2d] = popdata[idx2d] * rates[i]

        return fatgrid

    def getLossByShapes(
        self, mmidata, popdata, isodata, shapes, geodict, eventyear=None, gdpobj=None
    ):
        """Divide the losses calculated per grid cell into polygons that intersect with the grid.

        :param mmidata:
          Array of MMI values, dimensions (M,N).
        :param popdata:
          Array of population values, dimensions (M,N).
        :param isodata:
          Array of numeric country code values, dimensions (M,N).
        :param shapes:
          Sequence of GeoJSON-like polygons as returned from fiona.open().
        :param eventyear:
          4 digit event year, must be not None if loss type is economic.
        :param gdpobj:
          GDP object, containing per capita GDP data from all countries.
          Must not be None if calculating economic losses.
        :returns:
          Tuple of:
            1) modified sequence of polygons, including a new field "fatalities" or "dollars_lost".
            2) Total number of losses in all polygons.
        """
        lossgrid = self.getLossGrid(mmidata, popdata, isodata)
        polyshapes = []
        totloss = 0
        if self._loss_type == "fatality":
            fieldname = "fatalities"
        else:
            fieldname = "dollars_lost"
        for polyrec in shapes:
            polygon = shapely.geometry.shape(polyrec["geometry"])
            # overlay the polygon on top of a grid, turn polygon pixels to 1, non-polygon pixels to 0.
            tgrid = Grid2D.rasterizeFromGeometry(
                [polygon],
                geodict,
                fillValue=0,
                burnValue=1.0,
                attribute="value",
                mustContainCenter=True,
            )
            # get the indices of the polygon cells
            shapeidx = tgrid.getData() == 1.0
            # get the sum of those cells in the loss grid
            losses = np.nansum(lossgrid[shapeidx])
            polyrec["properties"][fieldname] = int(losses)
            polyshapes.append(polyrec)
            totloss += int(losses)

        return (polyshapes, totloss)



#!/usr/bin/env python

# stdlib imports
import logging
import pathlib
from datetime import timedelta

# third party imports
import numpy as np
import pandas as pd
from mapio.reader import get_file_geodict, read
from mapio.shake import ShakeGrid

# local imports
from esi_utils_pager.country import Country
from esi_utils_pager.growth import PopulationGrowth
from esi_utils_pager.pandas_container import PandasContainer

# constants indicating what values in urban/rural grid stand for
URBAN = 2
RURAL = 1

# constants for determining time of day
DAY_START_HOUR = 10
DAY_END_HOUR = 17
TRANSIT_START_HOUR_MORNING = 5
TRANSIT_END_HOUR_MORNING = 10
TRANSIT_START_HOUR_EVENING = 17
TRANSIT_END_HOUR_EVENING = 22
NIGHT_START_HOUR_EVENING = 22
NIGHT_END_HOUR_EVENING = 24
NIGHT_START_HOUR_MORNING = 0
NIGHT_END_HOUR_MORNING = 5

# US PAGER-specific country codes
EASTERN_US_CCODE = 903
WESTERN_US_CCODE = 904
CALIFORNIA_US_CCODE = 902
US_CCODE = 840

# time of day columns in casualty spreadsheets
TIMES = {"day": "CasualtyDay", "transit": "CasualtyDay", "night": "CasualtyNight"}


def add_dicts(dict1, dict2):
    """Sum the values of two dictionaries.

    :param dict1:
      Dictionary (possibly empty) containing keys with floating point values.
    :param dict2:
      Dictionary (possibly empty) containing keys with floating point values.
    :returns:
      New dictionary, with all unique keys from dict1/2, and the sum of the
      values from each key in both dictionaries.
      Example:
        dict1 = {'W1':60.0,'W2':75.0}
        dict2 = {'W1':40.0,'W2':25.0,'W3':100.0}
        add_dicts(dict1,dict2) => {'W1':100.0,'W2':100.0,'W3':100.0}
    """
    result = {}
    for key, value in dict1.items():
        if key not in result:
            result[key] = value
        else:
            result[key] += value
    for key, value in dict2.items():
        if key not in result:
            result[key] = value
        else:
            result[key] += value
    return result


def pop_dist(popi, workforce, time, dclass):
    """
    Calculate population distribution (residential, non-residential, and outdoor).
    :param popi:
      Total population.
    :param workforce:
      Pandas Series containing indices (WorkForceTotal,WorkForceAgricultural,
        WorkForceIndustrial,WorkForceServices)
    :param time: Time of day ('day','night','transit')
    :param dclass: Density class (URBAN,RURAL)
    :return: Tuple of (residential,non-residential,outdoor) population numbers.
    """
    fwf = workforce.WorkForceTotal
    fagr = workforce.WorkForceAgricultural
    f_ind = workforce.WorkForceIndustrial
    fser = workforce.WorkForceServices
    fnwf = 1 - fwf
    if dclass == URBAN:
        # DAY TIME
        FDRWF = 0.40
        FDRWFI = 0.01
        FDRWFS = 0.01
        FDRWFA = 0.01

        FDNRWFI = 0.89
        FDNRWFS = 0.89
        FDNRWFA = 0.34
        # EDUCATIONAL SERVICES FRACTION NOT ACCOUNTED IN WORKFORCE DIST.
        FDNRSCH = 0.25

        FDOWF = 0.35
        FDOWFI = 0.1
        FDOWFS = 0.1
        FDOWFA = 0.65

        # TRANSIT TIME
        FTRWF = 0.75
        FTRWFI = 0.20
        FTRWFS = 0.25
        FTRWFA = 0.45

        FTNRWFI = 0.25
        FTNRWFS = 0.25
        FTNRWFA = 0.01

        FTOWF = 0.25
        FTOWFI = 0.55
        FTOWFS = 0.50
        FTOWFA = 0.54

        # NIGHT TIME
        FNRWF = 0.999
        FNRWFI = 0.84
        FNRWFS = 0.89
        FNRWFA = 0.998

        FNNRWFI = 0.15
        FNNRWFS = 0.1
        FNNRWFA = 0.001

        FNOWF = 0.001
        FNOWFI = 0.01
        FNOWFS = 0.01
        FNOWFA = 0.001
    else:
        # RURAL
        # DAY TIME
        FDRWF = 0.40
        FDRWFI = 0.05
        FDRWFS = 0.05
        FDRWFA = 0.01

        FDNRWFI = 0.85
        FDNRWFS = 0.85
        FDNRWFA = 0.04
        # EDUCATIONAL SERVICES FRACTION NOT ACCOUNTED IN WORKFORCE DIST.
        FDNRSCH = 0.25

        FDOWF = 0.35
        FDOWFI = 0.1
        FDOWFS = 0.1
        FDOWFA = 0.95

        # TRANSIT TIME
        FTRWF = 0.80
        FTRWFI = 0.10
        FTRWFS = 0.15
        FTRWFA = 0.65

        FTNRWFI = 0.20
        FTNRWFS = 0.20
        FTNRWFA = 0.01

        FTOWF = 0.20
        FTOWFI = 0.70
        FTOWFS = 0.65
        FTOWFA = 0.34

        # NIGHT TIME
        FNRWF = 0.999
        FNRWFI = 0.89
        FNRWFS = 0.89
        FNRWFA = 0.998

        FNNRWFI = 0.1
        FNNRWFS = 0.1
        FNNRWFA = 0.001

        FNOWF = 0.001
        FNOWFI = 0.01
        FNOWFS = 0.01
        FNOWFA = 0.001

    if time == "day":
        respop = popi * (
            FDRWF * fnwf
            + FDRWFI * fwf * f_ind
            + FDRWFS * fwf * fser
            + FDRWFA * fwf * fagr
        )
        nrpop = popi * (
            FDNRWFI * fwf * f_ind
            + FDNRWFS * fwf * fser
            + FDNRWFA * fwf * fagr
            + FDNRSCH * fnwf
        )
        outpop = popi * (
            FDOWF * fnwf
            + FDOWFI * fwf * f_ind
            + FDOWFS * fwf * fser
            + FDOWFA * fwf * fagr
        )
    elif time == "transit":
        respop = popi * (
            FTRWF * fnwf
            + FTRWFI * fwf * f_ind
            + FTRWFS * fwf * fser
            + FTRWFA * fwf * fagr
        )
        nrpop = popi * (
            FTNRWFI * fwf * f_ind + FTNRWFS * fwf * fser + FTNRWFA * fwf * fagr
        )
        outpop = popi * (
            FTOWF * fnwf
            + FTOWFI * fwf * f_ind
            + FTOWFS * fwf * fser
            + FTOWFA * fwf * fagr
        )

    elif time == "night":
        respop = popi * (
            FNRWF * fnwf
            + FNRWFI * fwf * f_ind
            + FNRWFS * fwf * fser
            + FNRWFA * fwf * fagr
        )
        nrpop = popi * (
            FNNRWFI * fwf * f_ind + FNNRWFS * fwf * fser + FNNRWFA * fwf * fagr
        )
        outpop = popi * (
            FNOWF * fnwf
            + FNOWFI * fwf * f_ind
            + FNOWFS * fwf * fser
            + FNOWFA * fwf * fagr
        )

    respop = np.atleast_1d(np.squeeze(respop))
    nrpop = np.atleast_1d(np.squeeze(nrpop))
    outpop = np.atleast_1d(np.squeeze(outpop))
    return (respop, nrpop, outpop)


def load_panel_from_excel(excelfile):
    """Load a pandas Panel object by reading it from an Excel spreadsheet.

    :param excelfile:
      Path to Excel file.
    :returns:
      pandas Panel object.
    """
    paneldict = {}
    xl = pd.ExcelFile(excelfile)
    for sheet in xl.sheet_names:
        frame = pd.read_excel(excelfile, sheet_name=sheet)
        paneldict[sheet] = frame
    panel = pd.Panel(paneldict)
    return panel


def get_time_of_day(dtime, lon):
    """Determine time of day (one of 'day','night', or 'transit'), and event year and hour.

    :param dtime:
      Datetime object, representing event origin time in UTC.
    :param lon:
      Float longitude of earthquake origin.
    :returns:
      Tuple of time of day,local event year, and local event hour
    """
    # inputs datetime in utc, and local longitude
    toffset = lon / 15
    event_time = dtime + timedelta(0, toffset * 3600)
    event_year = event_time.year
    event_hour = event_time.hour
    timeofday = None
    if event_hour >= DAY_START_HOUR and event_hour < DAY_END_HOUR:
        timeofday = "day"

    transit1 = (
        event_hour >= TRANSIT_START_HOUR_MORNING
        and event_hour < TRANSIT_END_HOUR_MORNING
    )
    transit2 = (
        event_hour >= TRANSIT_START_HOUR_EVENING
        and event_hour < TRANSIT_END_HOUR_EVENING
    )
    if transit1 or transit2:
        timeofday = "transit"

    night1 = (
        event_hour >= NIGHT_START_HOUR_EVENING and event_hour <= NIGHT_END_HOUR_EVENING
    )
    night2 = (
        event_hour >= NIGHT_START_HOUR_MORNING and event_hour <= NIGHT_END_HOUR_MORNING
    )
    if night1 or night2:
        timeofday = "night"
    return (timeofday, event_year, event_hour)


class SemiEmpiricalFatality(object):
    def __init__(self, inventory, collapse, casualty, workforce, growth):
        """Create Semi-Empirical Fatality Model object.

        :param inventory:
          PandasContainer, containing DataFrames named:
           - 'BuildingTypes',
           - 'RuralNonResidential'
           - 'RuralResidential'
           - 'UrbanNonResidential'
           - 'UrbanResidential'

           where BuildingTypes is a Dataframe with columns:
             - Code: Building Code ('A','S','C1',etc.)
             - ShortDescription:  A short description of building type
                                  (i.e, 'adobe block').
             - OperationalDescription: A (sometimes) shorter description of
                                       building type.
             - LongDescription: Full description of building type,
                                i.e. 'High-Rise Ductile Reinforced Concrete Moment
                                Frame'
          All other Dataframes have columns:
            - CountryCode: Two letter ISO country code ('US',etc.)
            - CountryName: Name of country.
            - Building Codes, as listed and described in BuildingTypes Dataframe.

        :param collapse:
          PandasContainer, where first index is country (by two-letter country code),
            columns are:
            - BuildingCode:  Building code as above.
            - (IMT)_6.0 -> (IMT)_9.0 Columns with collapse rates at each (IMT) interval,
                           where (IMT) could be MMI,PGA,PGV,PSA1.0, etc.

        :param casualty:
          PandasContainer, where first index is country (by two-letter country code),
            columns are:
            - BuildingCode:  Building code as above.
            - CasualtyDay:   Casualty rate, given collapse, during the day.
            - CasualtyNight: Casualty rate, given collapse, during the night.

        :param workforce:
          PandasContainer consisting of a single dataframe, where rows are by country,
              and columns are:
            - CountryCode two letter ISO country code.
            - CountryCode name of country.
            - WorkforceTotal Fraction of the country population in the workforce.
            - WorkforceAgriculture Fraction of the total workforce employed in
              agriculture.
            - WorkforceIndustrial Fraction of the total workforce employed in industry.
            - WorkforceServices Fraction of the total workforce employed in services.
        :param growth:
          PopulationGrowth object.

        """
        self._inventory = inventory
        self._collapse = collapse
        self._casualty = casualty
        self._workforce = workforce
        self._popgrowth = growth
        self._country = Country()

    @classmethod
    def fromDefault(cls):
        homedir = pathlib.Path(__file__).parent  # where is this module?
        inventory_file = homedir / "data" / "semi_inventory.xlsx"
        collapse_file = homedir / "data" / "semi_collapse_mmi.xlsx"
        casualty_file = homedir / "data" / "semi_casualty.xlsx"
        workforce_file = homedir / "data" / "semi_workforce.xlsx"
        return cls.fromFiles(
            inventory_file, collapse_file, casualty_file, workforce_file
        )

    @classmethod
    def fromFiles(cls, inventory_file, collapse_file, casualty_file, workforce_file):
        """Create SemiEmpiricalFatality object from a number of input files.

        :param inventory_file:
          Excel file containing Semi-Empirical building inventory data.
          (described in __init__).
        :param collapse_file:
          Excel file containing Semi-Empirical collapse rate data  in an PandasContainer.
          (described in __init__).
        :param casualty_file:
          Excel file containing Semi-Empirical casualty rate data in an PandasContainer.
          (described in __init__).
        :param workforce_file:
          Excel file containing Semi-Empirical workforce data in an PandasContainer.
          (described in __init__).
        :param growth_file:
          Excel spreadsheet containing population growth rate data
          (described in PopulationGrowth.fromUNSpreadsheet()).
        :returns:
          SemiEmpiricalFatality object.
        """
        # turn the inventory,collapse, and casualty spreadsheets into Panels...
        inventory = PandasContainer(inventory_file)
        collapse = PandasContainer(collapse_file)
        casualty = PandasContainer(casualty_file)
        workforce = PandasContainer(workforce_file)
        # extract the one dataframe from the Panel
        workforce = workforce.getDataFrame("Workforce")
        workforce = workforce.set_index("CountryCode")

        # read the growth spreadsheet into a PopulationGrowth object...
        popgrowth = PopulationGrowth.fromDefault()

        return cls(inventory, collapse, casualty, workforce, popgrowth)

    def setGlobalFiles(self, popfile, popyear, urbanfile, isofile):
        """Set global files (pop,urban/rural, country code) for use of model with ShakeMaps.

        :param popfile:
          File name of population grid.
        :param popyear:
          Year population data was collected.
        :param urbanfile:
          File name of urban/rural grid (rural cells contain 1, urban cells contain 2).
        :param isofile:
          File name of numeric ISO country code grid.
        :returns:
          None
        """
        self._popfile = popfile
        self._popyear = popyear
        self._urbanfile = urbanfile
        self._isofile = isofile

    def getBuildingDesc(self, btype, desctype="short"):
        """Get a building description given a short building type code.

        :param btype:
          Short building type code ('A' (adobe), 'C' (reinforced concrete), etc.)
        :param desctype:
          A string, one of:
            - 'short': Very short descriptions ('adobe block')
            - 'operational': Short description, intended for use in automatically
               generated sentences about building types.
            - 'long': Most verbose description
               ('Adobe block (unbaked dried mud block) walls')

        :returns:
          Either a short, operational, or long description of building types.
        """
        bsheet = self._inventory.getDataFrame("BuildingTypes")
        bsheet = bsheet.set_index("Code")
        row = bsheet.loc[btype]
        if desctype == "short":
            return row["ShortDescription"]
        elif desctype == "operational":
            return row["OperationalDescription"]
        else:
            return row["LongDescription"]
        return None

    def getWorkforce(self, ccode):
        """Get the workforce data corresponding to a given country code.
        :param ccode:
          Two letter ISO country code.
        :returns:
          Pandas series containing Workforce data for given country
            (WorkForceTotal,WorkForceAgriculture,WorkForceIndustrial,WorkForceServices)
        """
        try:
            wforce = self._workforce.loc[ccode]
        except Exception:
            wforce = None
        return wforce

    def getCollapse(self, ccode, mmi, inventory):
        """Return the collapse rates for a given country,intensity, and inventory.

        :param ccode:
          Two letter ISO country code.
        :param mmi:
          MMI value (one of 6.0,6.5,7.0,7.5,8.0,8.5,9.0)
        :param inventory:
          Pandas Series containing an inventory for the given country.
        :returns:
          Pandas Series object containing the collapse rates for given building types,
          ccode, and MMI.
        """
        collapse_frame = self._collapse.getDataFrame(ccode)
        collapse_frame = collapse_frame.set_index("BuildingCode")
        collapse_frame = collapse_frame.loc[inventory.index]

        mmicol = f"MMI_{mmi:.1f}"
        collapse = collapse_frame[mmicol]
        return collapse

    def getFatalityRates(self, ccode, timeofday, inventory):
        """Return fatality rates for a given country, time of day, and inventory.

        :param ccode:
          Two-letter ISO country code.
        :param timeofday:
          One of 'day','transit', or 'night'.
        :param inventory:
          Pandas Series containing an inventory for the given country.
        :returns:
          Pandas Series object containing fatality rates for given country, time of day,
          and inventory.
        """
        fatalframe = self._casualty.getDataFrame(ccode)
        fatalframe = fatalframe.set_index("BuildingCode")
        timecol = TIMES[timeofday]
        idx = inventory.index
        fatrates = fatalframe.loc[idx][timecol]
        return fatrates

    def getInventories(self, ccode, density):
        """Return two pandas Series objects with urban or rural inventory for given country.

        :param ccode:
          Two-letter ISO country code.
        :param density:
          One of semimodel.URBAN (2) or semimodel.RURAL (1).
        :returns:
          Two Pandas Series: 1) Residential Inventory and 2) Non-Residential Inventory.
        """
        if density == URBAN:
            resinv = self._inventory.getDataFrame("UrbanResidential")
            nresinv = self._inventory.getDataFrame("UrbanNonResidential")
        else:
            resinv = self._inventory.getDataFrame("RuralResidential")
            nresinv = self._inventory.getDataFrame("RuralNonResidential")
        resinv = resinv.set_index("CountryCode")
        nresinv = nresinv.set_index("CountryCode")

        # we may be missing inventory for certain countries (Bonaire?). Return empty
        # series.
        if ccode not in resinv.index or ccode not in nresinv.index:
            return (pd.Series(), pd.Series())

        # pandas series of residential inventory
        resrow = resinv.loc[ccode]
        resrow = resrow.drop("CountryName")
        # pandas series of non-residential inventory
        nresrow = nresinv.loc[ccode]
        nresrow = nresrow.drop("CountryName")

        # now trim down the series to only include finite and non-zero values
        resrow = resrow[resrow.notnull()]
        resrow = resrow[resrow > 0]
        nresrow = nresrow[nresrow.notnull()]
        nresrow = nresrow[nresrow > 0]

        return (resrow, nresrow)

    def getLosses(self, shakefile):
        """Calculate number of fatalities using semi-empirical approach.

        :param shakefile:
          Path to a ShakeMap grid.xml file.
        :returns:
          Tuple of:
            1) Total number of fatalities
            2) Dictionary of residential fatalities per building type, per country.
            3) Dictionary of non-residential fatalities per building type, per country.
        """
        # get shakemap geodict
        shakedict = ShakeGrid.getFileGeoDict(shakefile, adjust="res")
        # get population geodict
        popdict = get_file_geodict(self._popfile)

        # get country code geodict
        isodict = get_file_geodict(self._isofile)

        # get urban grid geodict
        urbdict = get_file_geodict(self._urbanfile)

        # load all of the grids we need
        if popdict == shakedict == isodict == urbdict:
            # special case, probably for testing...
            shakegrid = ShakeGrid.load(shakefile, adjust="res")
            popgrid = read(self._popfile)
            isogrid = read(self._isofile)
            urbgrid = read(self._urbanfile)
        else:
            sampledict = popdict.getBoundsWithin(shakedict)
            shakegrid = ShakeGrid.load(
                shakefile,
                samplegeodict=sampledict,
                resample=True,
                method="linear",
                adjust="res",
            )
            popgrid = read(self._popfile, samplegeodict=sampledict, resample=False)
            isogrid = read(
                self._isofile,
                samplegeodict=sampledict,
                resample=True,
                method="nearest",
                doPadding=True,
                padValue=0,
            )
            urbgrid = read(
                self._urbanfile,
                samplegeodict=sampledict,
                resample=True,
                method="nearest",
                doPadding=True,
                padValue=RURAL,
            )

        # determine the local apparent time of day (based on longitude)
        edict = shakegrid.getEventDict()
        etime = edict["event_timestamp"]
        elon = edict["lon"]
        time_of_day, event_year, event_hour = get_time_of_day(etime, elon)

        # round off our MMI data to nearest 0.5 (5.5 should stay 5.5, 5.4
        # should become 5.5, 5.24 should become 5.0, etc.)
        mmidata = np.round(shakegrid.getLayer("mmi").getData() / 0.5) * 0.5

        # get arrays from our other grids
        popdata = popgrid.getData()
        isodata = isogrid.getData()
        urbdata = urbgrid.getData()

        # modify the population values for growth rate by country
        ucodes = np.unique(isodata[~np.isnan(isodata)])
        for ccode in ucodes:
            cidx = isodata == ccode
            popdata[cidx] = self._popgrowth.adjustPopulation(
                popdata[cidx], ccode, self._popyear, event_year
            )

        # create a dictionary containing indoor populations by building type (in cells
        # where MMI >= 6)
        # popbystruct = get_indoor_pop(mmidata,popdata,urbdata,isodata,time_of_day)

        # find all mmi values greater than 9, set them to 9
        mmidata[mmidata > 9.0] = 9.0

        # dictionary containers for sums of fatalities (res/nonres) by building type
        res_fatal_by_ccode = {}
        nonres_fatal_by_ccode = {}

        # fatality sum
        ntotal = 0

        # loop over countries
        ucodes = np.unique(isodata[~np.isnan(isodata)])
        for ucode in ucodes:
            if ucode == 0:
                continue
            res_fatal_by_btype = {}
            nonres_fatal_by_btype = {}

            cdict = self._country.getCountry(int(ucode))
            ccode = cdict["ISO2"]
            # get the workforce Series data for the current country
            wforce = self.getWorkforce(ccode)
            if wforce is None:
                logging.info("No workforce data for %s.  Skipping." % (cdict["Name"]))
                continue

            # loop over MMI values 6-9
            for mmi in np.arange(6, 9.5, 0.5):
                c1 = mmidata == mmi
                c2 = isodata == ucode
                if ucode > 900 and ucode != CALIFORNIA_US_CCODE:
                    ucode = US_CCODE
                for dclass in [URBAN, RURAL]:
                    c3 = urbdata == dclass

                    # get the population data in those cells at MMI, in country, and
                    # density class
                    # I think I want an AND condition here
                    popcells = popdata[c1 & c2 & c3]
                    if not len(popcells):
                        continue

                    # get the population distribution across residential,
                    # non-residential, and outdoor.
                    res, nonres, outside = pop_dist(
                        popcells, wforce, time_of_day, dclass
                    )

                    # get the inventory for urban residential
                    resrow, nresrow = self.getInventories(ccode, dclass)

                    # now multiply the residential/non-residential population through
                    # the inventory data
                    numres = len(resrow)
                    numnonres = len(nresrow)
                    resmat = np.reshape(resrow.values, (numres, 1)).astype(np.float32)
                    nresmat = np.reshape(nresrow.values, (numnonres, 1)).astype(
                        np.float32
                    )
                    popres = np.tile(res, (numres, 1))
                    popnonres = np.tile(nonres, (numnonres, 1))
                    popresbuilding = popres * resmat
                    popnonresbuilding = popnonres * nresmat

                    # now we have the residential and non-residental population
                    # distributed through the building types for each cell that matches
                    # MMI,country, and density criteria.
                    # popresbuilding rows are building types, columns are population
                    # cells

                    # next, we get the collapse rates for these buildings
                    # and multiply them by the population by building.
                    collapse_res = self.getCollapse(ccode, mmi, resrow)
                    collapse_nonres = self.getCollapse(ccode, mmi, nresrow)
                    resrates = np.reshape(
                        collapse_res.values.astype(np.float32), (numres, 1)
                    )
                    nonresrates = np.reshape(
                        collapse_nonres.values.astype(np.float32), (numnonres, 1)
                    )
                    rescollapse = popresbuilding * resrates
                    nonrescollapse = popnonresbuilding * nonresrates

                    # get the fatality rates given collapse by building type and
                    # multiply through the result of collapse*population per building
                    resfatalcol = self.getFatalityRates(ccode, time_of_day, resrow)
                    nonresfatalcol = self.getFatalityRates(ccode, time_of_day, nresrow)
                    resfatal = np.reshape(
                        resfatalcol.values.astype(np.float32), (numres, 1)
                    )
                    nonresfatal = np.reshape(
                        nonresfatalcol.values.astype(np.float32), (numnonres, 1)
                    )
                    resfat = rescollapse * resfatal
                    nonresfat = nonrescollapse * nonresfatal

                    # zero out the cells where fatalities are less than 1 or nan
                    try:
                        if len(resfat) and len(resfat[0]):
                            resfat[np.ma.masked_less(resfat, 1).mask] = 0.0
                    except Exception:
                        resfat[np.isnan(resfat)] = 0.0
                    try:
                        if len(nonresfat) and len(nonresfat[0]):
                            nonresfat[np.ma.masked_less(nonresfat, 1).mask] = 0.0
                    except Exception:
                        nonresfat[np.isnan(nonresfat)] = 0.0

                    # sum the fatalities per building through all cells
                    resfatbybuilding = np.nansum(resfat, axis=1)
                    nonresfatbybuilding = np.nansum(nonresfat, axis=1)
                    resfdict = dict(zip(resrow.index, resfatbybuilding.tolist()))
                    nonresfdict = dict(zip(nresrow.index, nonresfatbybuilding.tolist()))
                    res_fatal_by_btype = add_dicts(res_fatal_by_btype, resfdict)
                    nonres_fatal_by_btype = add_dicts(
                        nonres_fatal_by_btype, nonresfdict
                    )

            # add the fatalities by building type to the dictionary containing
            # fatalities by country
            res_fatal_by_ccode[ccode] = res_fatal_by_btype.copy()
            nonres_fatal_by_ccode[ccode] = nonres_fatal_by_btype.copy()

            # increment the total number of fatalities
            ntotal += int(
                sum(res_fatal_by_btype.values()) + sum(nonres_fatal_by_btype.values())
            )

        return (ntotal, res_fatal_by_ccode, nonres_fatal_by_ccode)



# stdlib imports
import logging
import pathlib
import sys
import warnings

# third party imports
import pandas as pd
import rasterio
from esi_utils_time.timeutils import LocalTime
from mapio.shake import getHeaderData
from rasterio.sample import sample_gen

# local imports
from esi_utils_pager.config import read_config
from esi_utils_pager.country import Country
from esi_utils_pager.econexposure import EconExposure
from esi_utils_pager.emploss import EmpiricalLoss
from esi_utils_pager.exposure import Exposure
from esi_utils_pager.semimodel import SemiEmpiricalFatality


def get_pop_year(config, event_year):
    # find the population data collected most closely to the event_year
    pop_year = None
    tmin = 10000000
    popfile = None
    for popdict in config["model_data"]["population_data"]:
        popyear = popdict["population_year"]
        popgrid = pathlib.Path(popdict["population_grid"])
        if not popgrid.is_file():
            print("Population grid file %s does not exist." % popgrid)
            sys.exit(1)
        if abs(popyear - event_year) < tmin:
            tmin = abs(popyear - event_year)
            pop_year = popyear
            popfile = popgrid

    return (pop_year, popfile)


def get_exposure(master_row, isofile, popfile, pop_year, gridfile):
    # Get exposure results

    expomodel = Exposure(popfile, pop_year, isofile)
    exposure = expomodel.calcExposure(gridfile)
    exp_rows = []
    for key, value in exposure.items():
        if key != "TotalExposure" and len(key) != 2:
            continue
        row = {}
        ccode = key
        if key == "TotalExposure":
            ccode = "Total"
        row["CountryCode"] = ccode
        headers = [f"MMI{i:02d}" for i in range(1, 11)]
        mmi_dict = dict(zip(headers, value))
        row.update(mmi_dict)
        exp_rows.append(row)

    expframe = pd.DataFrame(data=exp_rows)
    for key, value in master_row.items():
        expframe[key] = value
    allcols = expframe.columns
    remainder = set(allcols) - set(master_row.keys())
    newcols = list(master_row.keys()) + sorted(list(remainder))
    expframe = expframe[newcols].copy()
    maxmmi_array = []
    for _, row in expframe.iterrows():
        mmi_idx = row[headers] > 1000
        if not len(mmi_idx[mmi_idx].index):
            maxmmi_array.append(0)
            continue
        maxmmi_col = mmi_idx[mmi_idx].index[-1]
        maxmmi_array.append(headers.index(maxmmi_col) + 1)
    expframe["MaxMMI1000"] = maxmmi_array
    return (expframe, exposure)


def get_fatalities(expframe, exposure):
    fatmodel = EmpiricalLoss.fromDefaultFatality()
    fatdict = fatmodel.getLosses(exposure)
    fatdict["Total"] = fatdict.pop("TotalFatalities")
    fatframe = expframe.copy()
    fatframe["Fatalities"] = 0
    for key, value in fatdict.items():
        fatframe.loc[fatframe["CountryCode"] == key, "Fatalities"] = value
    return fatframe


def get_econ_losses(fatframe, popfile, popyear, isofile, gridfile):
    econexpmodel = EconExposure(popfile, popyear, isofile)
    ecomodel = EmpiricalLoss.fromDefaultEconomic()
    econexposure = econexpmodel.calcExposure(gridfile)
    rows = []
    for key, value in econexposure.items():
        if key != "TotalEconomicExposure" and len(key) != 2:
            continue
        row = {}
        ccode = key
        if key == "TotalEconomicExposure":
            ccode = "Total"
        row["CountryCode"] = ccode
        headers = [f"EconMMI{i:02d}" for i in range(1, 11)]
        mmi_dict = dict(zip(headers, value))
        row.update(mmi_dict)
        rows.append(row)
    econframe = pd.DataFrame(data=rows)
    econframe = pd.merge(fatframe, econframe, on="CountryCode")
    ecodict = ecomodel.getLosses(econexposure)
    ecodict["Total"] = ecodict.pop("TotalDollars")
    econframe["Dollars"] = 0
    for key, value in ecodict.items():
        econframe.loc[econframe["CountryCode"] == key, "Dollars"] = value
    return econframe


def get_local_time(etime, timezone_file, lat, lon):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        ltime = LocalTime(timezone_file, etime, lat, lon)
        localtime = ltime.getLocalTime()
    return localtime


def get_semi_losses(gridfile, popfile, popyear, isofile, urbanfile):
    semimodel = SemiEmpiricalFatality.fromDefault()
    semimodel.setGlobalFiles(popfile, popyear, urbanfile, isofile)
    # Tuple of:
    #         1) Total number of fatalities
    #         2) Dictionary of residential fatalities per building type, per country.
    #         3) Dictionary of non-residential fatalities per building type, per country.
    semifat, resfat, nonresfat = semimodel.getLosses(gridfile)
    return (semifat, resfat, nonresfat)


def calc_pager_event(gridfile, config, run_semi):
    # get all the basic event information and print it, if requested
    shake_tuple = getHeaderData(gridfile)
    local_time = get_local_time(
        shake_tuple[1]["event_timestamp"],
        config["model_data"]["timezones_file"],
        shake_tuple[1]["lat"],
        shake_tuple[1]["lon"],
    )
    master_row = {}
    master_row["EventID"] = shake_tuple[1]["event_id"]
    master_row["Time"] = shake_tuple[1]["event_timestamp"]
    master_row["LocalTime"] = local_time
    master_row["Latitude"] = shake_tuple[1]["lat"]
    master_row["Longitude"] = shake_tuple[1]["lon"]
    master_row["Depth"] = shake_tuple[1]["depth"]
    master_row["Magnitude"] = shake_tuple[1]["magnitude"]
    master_row["Location"] = shake_tuple[1]["event_description"]
    event_year = shake_tuple[1]["event_timestamp"].year
    isofile = config["model_data"]["country_grid"]
    country = Country()
    with rasterio.open(isofile, "r") as dataset:
        xy = [(shake_tuple[1]["lon"], shake_tuple[1]["lat"])]
        isocode = list(sample_gen(dataset, xy))[0][0]
        ccode = country.getCountry(isocode)["ISO2"]
    master_row["EpicentralCountryCode"] = ccode
    pop_year, popfile = get_pop_year(config, event_year)
    expframe, exposure = get_exposure(master_row, isofile, popfile, pop_year, gridfile)
    fatframe = get_fatalities(expframe, exposure)
    ecoframe = get_econ_losses(fatframe, popfile, pop_year, isofile, gridfile)

    if run_semi:
        urbanfile = config["model_data"]["urban_rural_grid"]
        semifat, resfat, nonresfat = get_semi_losses(
            gridfile, popfile, pop_year, isofile, urbanfile
        )
        resframe = pd.DataFrame(resfat).transpose()
        nonresframe = pd.DataFrame(nonresfat).transpose()
        final_frame = resframe.add(nonresframe)
        final_frame = final_frame.round()
        ny, nx = final_frame.shape
        if nx > 0 and ny > 0:
            ftotal = final_frame.sum(axis="rows")
            final_frame.loc["Total"] = ftotal
            btotal = final_frame.sum(axis="columns")
            final_frame["TotalSemiFatalities"] = btotal
            ecoframe = ecoframe.join(final_frame, on=["CountryCode"])

    return ecoframe


def calc_pager_events(gridfolder, verbose, run_semi):
    gridfolder = pathlib.Path(gridfolder)
    # read config file
    config = read_config()
    # Make sure grid.xml file exists
    if not gridfolder.is_dir():
        print(f"ShakeMap Grid folder {gridfolder} does not exist.")
        sys.exit(1)

    gridfiles = gridfolder.glob("**/*grid.xml")
    dataframes = []
    for gridfile in gridfiles:
        if verbose:
            logging.info(f"Parsing {gridfile}...")
        dataframe = calc_pager_event(gridfile, config, run_semi)
        dataframes.append(dataframe)

    dataframe = pd.concat(dataframes)
    if run_semi:
        # make TotalSemiFatalities column the last one, if it exists
        fatcol = dataframe.pop("TotalSemiFatalities")
        ncols = len(dataframe.columns)
        dataframe.insert(ncols, "TotalSemiFatalities", fatcol)
    return dataframe



import pandas as pd


class PandasContainer(object):
    def __init__(self, excelfile):
        # excel file with multiple tabs
        self._excelfile = excelfile
        self._excelfile = pd.ExcelFile(excelfile)

    def getDataFrame(self, name):
        if name not in self._excelfile.sheet_names:
            raise LookupError((f"Dataframe {name} not in {self._excelfile}"))
        return self._excelfile.parse(name)
